'''io/array.py'''

import numpy as np

from . import base

class ArrayReader(base.Reader):
    """
    A audiotsm2.io.base.Reader allowing to use numpy.ndarray as
    input of a audiotsm2.base.tsm.TSM object.

    a matrix of shape (m, n), with m the number of channels and n
    the length of the buffer, where the samples will be read.
    """
    def __init__(self, data):
        self._data = data

    @property
    def channels(self):
        return self._data.shape[0]

    @property
    def empty(self):
        return self._data.shape[1] == 0

    def read(self, buffer):
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of channels")

        # Number of samples to read
        n = min(buffer.shape[1], self._data.shape[1])
        np.copyto(buffer[:, :n], self._data[:, :n])

        # Remove the samples that were read
        self._data = self._data[:, n:]

        return n

    def skip(self, n):
        if n > self._data.shape[1]:
            n = self._data.shape[1]

        self._data = self._data[:, n:]

        return n


class ArrayWriter(base.Writer):
    """
    A audiotsm2.io.base.Writer allowing to get the output of a
    audiotsm2.base.tsm.TSM object as a numpy.ndarray.

    Writing to an audiotsm2.io.array.ArrayWriter will add the data at
    the end of the audiotsm2.io.array.ArrayWriter.data attribute.
    """

    def __init__(self, channels):
        self._channels = channels
        self._data = []

    @property
    def channels(self):
        return self._channels

    def write(self, buffer):
        if buffer.shape[0] != self._channels:
            raise ValueError("the buffer should have the same number of "
                             "channels as the ArrayWriter")

        self._data.append(np.copy(buffer))

        return buffer.shape[1]

    @property
    def data(self):
        """
        A numpy.ndarray of shape (m, n), with m the number of channels and n the
        length of the data, where the samples have written.
        """
        if not self._data:
            return np.ndarray((self._channels, 0), dtype=np.float32)

        data = np.concatenate(self._data, axis=1)
        self._data = [data]

        return data


class FixedArrayWriter(base.Writer):
    """
    A audiotsm2.io.base.Writer allowing to use
    numpy.ndarray as output of a TSM object.

    Contrary to an audiotsm2.io.array.ArrayWriter, a
    audiotsm2.io.array.FixedArrayWriter takes the buffer in which the
    data will be written as a parameter of its constructor. The buffer is of
    fixed size, and it will not be possible to write more samples to the
    audiotsm2.io.array.FixedArrayWriter than the buffer can contain.
    """
    def __init__(self, data):
        self._data = data

    @property
    def channels(self):
        return self._data.shape[0]

    def write(self, buffer):
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of "
                             "channels")

        # Number of samples to write
        n = min(buffer.shape[1], self._data.shape[1])
        np.copyto(self._data[:, :n], buffer[:, :n])

        self._data = self._data[:, n:]

        return n



'''base/__init__.py'''

"""
This module provides base classes for the implementation of time-scale
modification procedures.
"""

from .tsm import TSM
from .analysis_synthesis import AnalysisSynthesisTSM, Converter


'''io/__init__.py'''

"""
audiotsm.base.tsm.TSM objects use audiotsm.io.base.Reader
objects as input and audiotsm.io.base.Writer objects as output.

The audiotsm.io package provides Readers and Writers allowing to use
numpy arrays or wav files as input or output of a
audiotsm.base.tsm.TSM, to play the output in real-time, as well as
base classes to implement your own Readers and Writers.
"""


# -*- coding: utf-8 -*-

class Reader(object):
    @property
    def channels(self):
        raise NotImplementedError()

    @property
    def empty(self):
        raise NotImplementedError()


class Writer(object):
    @property
    def channels(self):
        raise NotImplementedError()


# -*- coding: utf-8 -*-

"""
The audiotsm.gstreamer.phasevocoder module implements an audio filter
allowing to use the phase vocoder procedure with gstreamer.
"""

import gi
gi.require_version('Gst', '1.0')

from gi.repository import GObject, Gst
from audiotsm2 import phasevocoder
from .base import GstTSM

class PhaseVocoder(GstTSM):
    """Phase vocoder gstreamer audio filter."""
    __gstmetadata__ = (
        'Phase vocoder time-scale modification', 'Transform',
        'Change the speed of an audio stream with the phase vocoder procedure',
        'Muges'
    )

    plugin_name = "audiotsm-phase-vocoder"
    """The plugin name, to be used in ``Gst.ElementFactory.make``."""

    frame_length = GObject.Property(type=int, default=-1,
                                    flags=GObject.ParamFlags.WRITABLE)
    """The length of the frames.

    This is a write-only attribute, that will only take effect the next time
    the audio filter is setup (usually on the next song)."""

    synthesis_hop = GObject.Property(type=int, default=-1,
                                     flags=GObject.ParamFlags.WRITABLE)
    """The number of samples between two consecutive synthesis frames.

    This is a write-only attribute, that will only take effect the next time
    the audio filter is setup (usually on the next song)."""

    phase_locking = GObject.Property(type=int, default=-1,
                                     flags=GObject.ParamFlags.WRITABLE)
    """The phase locking strategy.

    This is a write-only attribute, that will only take effect the next time
    the audio filter is setup (usually on the next song)."""

    def create_tsm(self, channels):
        parameters = {}
        if self.frame_length > 0:
            parameters['frame_length'] = self.frame_length
        if self.synthesis_hop > 0:
            parameters['synthesis_hop'] = self.synthesis_hop
        if self.phase_locking >= 0:
            parameters['phase_locking'] = self.phase_locking

        return phasevocoder(channels, **parameters)


PhaseVocoder.register()

# Register the plugin to make it usable outside python, e.g. with the
# following commands (this does not seem to work):
# export GST_PLUGIN_PATH=$PWD/audiotsm/gstreamer/
# gst-launch-1.0 fakesrc num-buffers=10 ! audiotsm-phase-vocoder ! fakesink
_PV_TYPE = GObject.type_register(PhaseVocoder)
__gstelementfactory__ = (PhaseVocoder.plugin_name, Gst.Rank.NONE, _PV_TYPE)



# -*- coding: utf-8 -*-

"""
The audiotsm2.base.tsm module provides an abstract class for real-time
audio time-scale modification procedures.
"""

class TSM(object):
    """An abstract class for real-time audio time-scale modification
    procedures.
    """

    def clear(self):
        raise NotImplementedError

    def flush_to(self, writer):
        raise NotImplementedError

    def get_max_output_length(self, input_length):
        raise NotImplementedError

    def read_from(self, reader):
        raise NotImplementedError

    def run(self, reader, writer, flush=True):
        """
        Runs the TSM procedure on the content of reader and writes the output to writer.
        """
        finished = False
        while not (finished and reader.empty):
            self.read_from(reader)
            _, finished = self.write_to(writer)

        if flush:
            finished = False
            while not finished:
                _, finished = self.flush_to(writer)

            self.clear()

    def set_speed(self, speed):
        raise NotImplementedError

    def write_to(self, writer):
        raise NotImplementedError
